use alloc::string::String;
use log::{debug, error, info};
use uefi::prelude::*;
use uefi::proto::network::pxe::{BaseCode, UdpOpFlags};
use uefi::proto::network::IpAddress;
use uefi::table::boot::{ScopedProtocol, SearchType};
use uefi::{Error, Identify, Result};
extern crate alloc;
use crate::config::Config;
use uefi::data_types::CStr16;
use uefi::fs::{Path, PathBuf};

fn split_arg(raw_val: &str) -> Option<(&str, &str)> {
    let mut splitted = raw_val.split("=");
    let arg_name = match splitted.next() {
        None => return None,
        Some(v) => v,
    };
    let arg_value = match splitted.next() {
        None => return None,
        Some(v) => v,
    };
    return Some((arg_name, arg_value));
}

fn get_boot_path<'a>(msg: String) -> Result<PathBuf, &'a str> {
    debug!("Splitting message");

    // a message has the following format: "ACTION arg1=val1,arg2=val2"
    let mut splitted_msg_iter = msg.split_ascii_whitespace();

    debug!("Checking if message is a valid BOOT_ACCEPT message");
    match splitted_msg_iter.next() {
        None => return Err(Error::new(Status::UNSUPPORTED, "empty message")),
        Some(action) => match action {
            "BOOT_ACCEPT" => (),
            "BOOT_DENY" => return Err(Error::new(Status::UNSUPPORTED, "boot refused")),
            _ => return Err(Error::new(Status::UNSUPPORTED, "unexpected action")),
        },
    }

    debug!("Parsing message arguments");
    match splitted_msg_iter.next() {
        None => return Err(Error::new(Status::UNSUPPORTED, "empty list of arguments")),
        Some(args_list) => {
            // split all the arguments in a comma separated list
            for arg in args_list.split(",") {
                match split_arg(arg) {
                    None => debug!("unexpected format for argument {}", arg),
                    Some((arg_name, arg_value)) => match arg_name {
                        "id" => debug!("found id {}", arg_value),
                        "efi_app" => {
                            debug!("found efi app {}", arg_value);
                            let mut buf = [0; 128];
                            let path_raw = CStr16::from_str_with_buf(arg_value, &mut buf)
                                .expect("invalid str format");
                            let p = Path::new(path_raw);
                            let mut res = PathBuf::new();
                            res.push(p);
                            return Ok(res);
                        }
                        _ => debug!("unexpected argument {} with value {}", arg_name, arg_value),
                    },
                }
            }
        }
    }
    Err(Error::new(Status::UNSUPPORTED, "no efi app defined"))
}

fn read_mcast_response<'a>(base_client: &mut ScopedProtocol<'_, BaseCode>) -> Result<PathBuf> {
    let mut buf: [u8; 200] = [0; 200];

    let n = base_client
        .udp_read(
            UdpOpFlags::ANY_DEST_IP
                | UdpOpFlags::ANY_DEST_PORT
                | UdpOpFlags::ANY_SRC_IP
                | UdpOpFlags::ANY_SRC_PORT,
            None,
            None,
            None,
            None,
            None,
            &mut buf,
        )
        .unwrap();

    info!("Read {} bytes", n);
    let received_msg = String::from_utf8(buf[..n].to_vec()).unwrap();
    info!("Received message {}", received_msg);
    match get_boot_path(received_msg) {
        Ok(r) => Ok(r),
        Err(e) => {
            error!("failed to get boot path: {}", e.data());
            Err(e.to_err_without_payload())
        }
    }
}

pub fn send_mcast<'a>(boot_service: &BootServices, conf: &Config) -> Result<PathBuf> {
    debug!("Initializing base bootservice client handle");
    let base_client_handle = *boot_service
        .locate_handle_buffer(SearchType::ByProtocol(&BaseCode::GUID))?
        .first()
        .expect("DevicePathToText is missing");

    let mut base_client = boot_service.open_protocol_exclusive::<BaseCode>(base_client_handle)?;

    debug!("Initializing source IP address");
    let src_bytes: [u8; 16] = [
        0x20, 0x01, 0x0d, 0xb8, 0x00, 0x00, 0x00, 0x00, 0x02, 0x16, 0x3e, 0xff, 0xfe, 0xc2, 0x16,
        0xb7,
    ];
    let src_address: IpAddress = IpAddress::new_v6(src_bytes);

    let mut req_msg = String::from("BOOT_REQUEST id=");
    req_msg.push_str(&conf.client_uuid);

    info!("Starting PXE client");
    debug!("Starting PXE client");
    base_client.start(true)?;

    debug!("Setting IP address for PXE client");
    base_client.set_station_ip(Some(&src_address), None)?;

    info!("Sending UDP multicast message");
    base_client.udp_write(
        UdpOpFlags::MAY_FRAGMENT,
        &conf.multicast_group,
        conf.multicast_port,
        None,
        Some(&src_address),
        None,
        None,
        req_msg.as_bytes(),
    )?;
    let res = read_mcast_response(&mut base_client);

    info!("Stopping PXE client");
    base_client.stop()?;

    info!("PXE client successfully executed");
    res
}
